#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <auth.h>
#include <mp.h>
#include <libsec.h>
#include <bio.h>
#include <ndb.h>

#include "httpd.h"
#include "httpsrv.h"
#include "misc.h"

#define setenv(var, value) value?putenv(var, value):putenv(var, "")

typedef struct Handle Handle;
struct Handle
{
//	char *mimetype;
	char **argv;
	int	hctl;
};

typedef struct ChunkBuf ChunkBuf;
struct ChunkBuf
{
	uchar	buf[Hsize+32];
	int	size;
	ChunkBuf *next;
};

/*	basic constant	*/
const char *serversoft="Pegasus/2.2 (Plan9)";
const char *HTTPLOG = "http";
static const char *httpdconf = "/sys/lib/httpd.conf";
static const char *httpport="80";
static const char *httpsport="443";
static const char *netdir = "/net";

/*	default options	*/
static char *webbase = "/usr/web";
static char *namespace = "/lib/namespace.httpd";
static char *rewrite = "/sys/lib/httpd.rewrite"; // system rewrite-file
char *port = "80"; // replaced later by really accepted port. used in log.c
static int	server = 0; // 0: runs under the control of listen. 1: runs in server mode.
static int mntok=0;
static char *runas="none";
static int usermode = 0; // 1: httpd service by the user who invoked httpd
static int monmode = 0; // 1: not run as daemon; monitored by `mon'
static int noweb = 0; // 0: allows $home/web as HomePage if present
int allowbasic = 0; // 1: allow basic authentication

/*	the followings may be required tuning for your server
*	and can be reset in configuration file "/sys/lib/httpd.conf"	*/
static int parsetimeout = 15; // timeout to parse header (in unit of sec)
static int waittimeout1 = 15; // wait timeout for a non-authenticated client (in unit of sec)
static int waittimeout2 = 900; // wait timeout for an authenticated client (in unit of sec)
static int cgitimeout = 5; // timeout for CGI (in unit of sec)
static int posttimeout = 15*60; // timeout for post data (in unit of sec)
static int maxpost = 10; // in unit of MB. post data size limit that must less than 512MB(comes from ramfs limit)
static int connectlimit = 300; // limit of total open tcp connection
static int maxconnect = 20; // max tcp connection by the IP
static int obstime = 3; // observation time to detect burst access (in unit of sec)
static int maxaccess = 20; // allowed max access in obstime
int lockouttime = 60*3; // lockout time for maxconnect and maxaccess (in unit of sec)
static int contmax = 100; // max persistent continuation count for safty

/*	basic constant that are given later	*/
static char *myip; //  server IP. 202.250.160.40 for example
static char *myname; // server name. example: ar in ar.aichi-u.ac.jp
static char *padom; // parent dom. example: aichi-u.ac.jp
static char *mydom; //	myname + "." + padom
static char *scheme; // http or https
char *uriconnection; // "http://ar.aichi-u.ac.jp/" ...  # scheme://host
static char *hostroot; // for example /usr/bob/www; real host's webroot

/* used internally for control */
int authenticated; /* 1 if authenticated. refered by send.c */
static int waittimeout;
static int cpid; // pid of CGI
static int mnt; // namespace configuration is finished
static int cont; // current persistent connection continuation count
static int notestatus; // alarm note to CGI is received
static char *webmnt; // to be "/usr/web/mnt"
static int binding; // 0: main document, 1: virtual host doc, 2: user's doc
static char *certificate;
static uchar *cert;
static int certlen;
static int ntossed; // current tossed count that controls redirection loop in persitent connection.
Db *cgih; // cgi headers that comes from cgi output, and from mimetype of CGI handler
Db *errh; // http header that should pass to fail message.

static char *pathinfo;
static char *scriptname;



/*	Note on cont
*	cont is used as a flag so that we can detect re-request using Keep-Alive
*	if (cont > 0)
*		re-request due to Kepp-Alive
*		so we should not reconfigure our namespace
*		"cont" means "continued"
*/

static	char *csquery(char*, char*, char*);
static	char *sysdom(char *name);
static	char *pardom(void);
static	void		becomenone(void);
static	int		nsconfig(HConnect *c, char *webbase);
static	void		dolisten(char*);
static	int		doreq(HConnect*);
static	int		send(HConnect*);
static 	void		doservice(char *ndir);
static	int		acceptpost(HConnect *c, int timeout);
static 	int		execute(HConnect *c, char **argv, int hctl);
static	int		hparse2env(HConnect *c, char *head);
static 	int		notefun(void *a, char *msg);
static 	int	execsafely(HConnect *c, char *path, char **argv, int hctl);
static	int		merge(char *mnt, char *root);
static	int	ipcheck(HConnect *c, char *file, char *ip, char *path);
static 	char*	redirect(HConnect *c, char *file, char *path, char **webroot);
static 	Handle*	handler(HConnect *c, char *file, char *path);
static	void		setcgienv(HConnect *c);
static 	char *	hcat2(HConnect *c, char *a1, char *a2);
static	char *	hcat3(HConnect *c, char *a1, char *a2, char *a3);
static	int		ramfs(HConnect *c);
static 	int		nomount(HConnect *c);
char *	addoptions(HConnect *c, char *uri);
static	int		getresheader(HConnect *c, char *status, char **resheader);
int	creject(int ctl, char *ndir);
int	httpstatus(char *status);

char *usage=
	"usage: httpd [-Mmsuw] [-t timeout] [-p port] [-c certificate] [-n namespace] [-R rewrite]  [[httpd] dummy ..  incalldir]\n";

void
main(int argc, char **argv)
{
	char buf[256];
	char *incalldir = nil;
	fmtinstall('D', hdatefmt);
	fmtinstall('H', httpfmt);
	fmtinstall('U', hurlfmt);
	if(readconfig(httpdconf) < 0){
		syslog(0, HTTPLOG, "# error in reading config # %r");
		exits("readconfig");
	}
	setstrconfig(&myname,"myname");
	setstrconfig(&webbase,"base");
	setstrconfig(&namespace,"namespace");
	setstrconfig(&rewrite,"rewrite");

	setintconfig(&allowbasic, "allowbasic");
	setintconfig(&waittimeout1, "waittimeout1");
	setintconfig(&waittimeout2, "waittimeout2");
	setintconfig(&parsetimeout, "parsetimeout");
	setintconfig(&cgitimeout, "cgitimeout");
	setintconfig(&posttimeout, "posttimeout");
	setintconfig(&maxpost, "maxpost");
	maxpost *= 1024*1024;
	setintconfig(&connectlimit, "connectlimit");
	setintconfig(&maxconnect, "maxconnect");
	setintconfig(&obstime, "obstime");
	setintconfig(&maxaccess,"maxaccess");
	setintconfig(&lockouttime,"lockouttime");
	setintconfig(&contmax,"contmax");
	
	waittimeout = waittimeout1;

	ARGBEGIN{
	case 'n':
		namespace = ARGF();
		break;
	case 'p':
		port = ARGF(); /* http */
		break;
	case 'c':
		certificate = ARGF();
		cert = readcert(certificate, &certlen);
		if(cert == nil)
			sysfatal("reading certificate: %r");
		break;
	case 't':
		cgitimeout = atoi(ARGF());
		break;
	case 'R':
		rewrite = ARGF();
		break;
	case 's':
		server = 1;
		break;
	case 'm':
		mntok = 1;
		break;
	case 'u':
		usermode = 1;
		break;
	case 'w':
		noweb = 1;
		break;
	case 'M':
		monmode = 1;
		break;
	default:
		sysfatal(usage);
		break;
	}ARGEND

	if(argc > 0)
		incalldir = argv[argc-1];
	if((server && incalldir) || (!server && !incalldir))
		sysfatal("# error: inconsistent use of -s and incalldir: %s", usage);
	if(!server && monmode)
		sysfatal("# error: inconsistent use of -M: %s", usage);

	if(webbase[0] != '/')
		sysfatal("webbase must begin with '/'");
	
	snprint(buf, sizeof buf, "%s/mnt", webbase);
	webmnt = estrdup(buf);

	if(namespace[0] != '/')
		sysfatal("namespace must begin with `/'");

	if(rewrite[0] != '/')
		sysfatal("rewrite must begin with `/'");

	padom = pardom(); // parent dom; aichi-u.ac.jp in ar.aichi-u.ac.jp
	if(padom == nil) // something wrong
		sysfatal("pardom() returned nil");
	if(myname == nil)
		myname = sysname();
	snprint(buf, sizeof buf, "%s.%s", myname, padom);
	mydom = estrdup(buf);
	scheme = cert?"https":"http";
	if(!server){
		if(usermode)
			sysfatal("inconsistent options");
		time(nil);
		contentinit();
		doservice(incalldir);
		exits(nil);
	}

	if(!monmode){
		switch(rfork(RFNOTEG|RFPROC|RFFDG|RFNAMEG)) {
		case -1:
			sysfatal("fork");
		case 0:
			break;
		default:
			exits(nil);
		}
	}

	/*
	 * open all files we might need before castrating namespace
	 */
	time(nil);
	syslog(0, HTTPLOG, nil);

	contentinit();

	runas = getuser();
	if(!usermode && (strcmp(runas, "none") != 0)){
		runas = "none";
		if(!certificate)
			becomenone();
	}
	dolisten(netmkaddr("*", "tcp", port));
	exits(nil);
}

static void
becomenone(void)
{
	int fd;

	fd = open("#c/user", OWRITE);
	if(fd < 0 || write(fd, "none", strlen("none")) < 0)
		sysfatal("can't become none");
	close(fd);
	if(newns("none", nil) < 0)
		sysfatal("can't build normal namespace");
}

static HConnect*
mkconnect(void)
{
	HConnect *c;

	c = ezalloc(sizeof(HConnect));
	c->hpos = c->header;
	c->hstop = c->header;
	c->replog = 0;
	return c;
}

static HSPriv*
mkhspriv(void)
{
	HSPriv *p;

	p = ezalloc(sizeof(HSPriv));
	return p;
}

static void
dolisten(char *address)
{
	char ndir[NETPATHLEN], dir[NETPATHLEN], *p;
	int n, m, k, t, t0, ctl, nctl, data, spotchk;
	TLSconn conn;
	char		netdirb[256];
	char *ip;

	spotchk = 0;
	if(cert)
		syslog(0, HTTPLOG, "httpd starting on %s using %s", address, certificate);
	else
		syslog(0, HTTPLOG, "httpd starting on %s", address);
	ctl = announce(address, dir);
	if(ctl < 0){
		syslog(0, HTTPLOG, "error: can't announce on %s: %r", address);
		return;
	}
	//syslog(0,HTTPLOG,"%s %s", address, dir);
	/* 	then
	*	address is "tcp!*!80"
	*	dir is " /net/tcp/23"
	*	term% netstat -n
	*	tcp  23   web        Listen       80         0          ::
	*/
	strcpy(netdirb, dir);
	p = nil;
	if(netdir[0] == '/'){
		p = strchr(netdirb+1, '/');
		if(p != nil)
			*p = '\0';
	}
	if(p == nil)
		strcpy(netdirb, "/net");
	netdir = netdirb; // this is "/net"

	t0 = time(nil);
	k = 0;
	for(;;){
		/*
		 *  wait for a call (or an error)
		 */
		nctl = listen(dir, ndir);
		/*	example:
		*	dir="/net/tcp/23"	# listenning using this directory
		*	ndir="/net/tcp/33"	# new connection
		*	dir is a listenning directory
		*	ndir is the assigned directory
		*	nctl is opened in listen() by open("/net/tcp/23/listen", ORDWR)
		*	and should be closed in signle connection
		*	listen() reads "/net/tcp/23/listen" to find ndir
		*	listen() returns open fd of "/net/tcp/23/listen"
		*	Ref: /sys/src/libc/9sys/announce.c		*/
		if(nctl < 0){
			syslog(0, HTTPLOG, "error: can't listen on %s: %r", address);
			syslog(0, HTTPLOG, "error+ ctls = %d", ctl);
			return;
		}

		/*	Some protection against DoS attack that intends to make
		*	huge numbers of connected port.
		*	Note that persistent connection loop cannot be detected here.
		*	Be careful not to reject requests from proxy server.	*/

		ip = remoteaddr(ndir);
		puthlist(ip);
		//syslog(0,HTTPLOG,"%s call %s", ip, port);
		if(black(ip)){
			creject(ctl, ndir);
			close(nctl);
			continue;
		}

		/*
		*	int   accept(int ctl, char *dir);
		*	int   reject(int ctl, char *dir, char *cause);
		*	accept() accepts a call received by listen(),
		*	while reject() refuses the call because of cause. 
		*	Ref: dial(2)
		*/
		t = time(nil);
		//syslog(0,HTTPLOG,"time: %d %d %d %d %d", t, t0, k, obstime,maxaccess);
		/*	time(nil) returns local clock in starting machine.
		*	however in few 10 seconds it might change to the time based on NTP.
		*	then time returned by time(nil) will jump discontinuously.
		*	the condition t >= t0 is added to protect the anomaly that comes
		*	from this phenomenon	*/
		if(t >= t0 && t - t0 < obstime)
			k++;
		else{
			k = 0;
			t0 = t;
		}
		if(k > maxaccess){
			/*	The condition is system dependent,
			*	therefore required tuning.	*/
			creject(ctl, ndir);
			close(nctl);
			ip = arrest(obstime); // detect the ip that causes busy
			if(ip){
				putblist(ip); // and throw that into blacklist
				syslog(0,"blacklist", "%s %s # maxaccess", port, ip);
			}
			else // should not happen
				syslog(0,"blacklist", "%s unknown # maxaccess", port);
			continue;
		}
	
		n = nconnect(ip, &m);
		if( n > maxconnect){
			syslog(0,"blacklist", "%s %s # maxconnect", port, ip);
			putblist(ip);// throw the ip into blacklist
			creject(ctl, ndir);
			close(nctl);
			continue;
		}
		if( m > connectlimit){
			creject(ctl, ndir);
			close(nctl);
			continue;
		}

		/*
		 *  start a process for the service
		 */
		switch(rfork(RFFDG|RFPROC|RFNOWAIT|RFNAMEG)){
		case -1:
			close(nctl);
			continue;
		case 0:
			/*
			 *  see if we know the service requested
			*
			*	What is the first argument of accept() should be ?
			*	man dial(2) says, it is nctl
			*	however /sys/src/cmd/ip/httpd/httpd.c says, ctl
			 */
			data = accept(ctl, ndir);
			if(data < 0){
				syslog(0, HTTPLOG, "error: can't open %s/data: %r", ndir);
				exits(nil);
			}
			if(cert){
				memset(&conn, 0, sizeof(conn));
				conn.cert = cert;
				conn.certlen = certlen;
				data = tlsServer(data, &conn);
				if(data < 0){// added by Kenar
					syslog(0, HTTPLOG, "error: tlsServer certificate=%s: %r", certificate);
					exits(nil);
				}
			}
			dup(data, 0);
			dup(data, 1);
			dup(data, 2);
			close(data);
			close(ctl);
			close(nctl);

			if(!usermode && certificate)
				becomenone();
			doservice(ndir);

			exits(nil);
		default:
			close(nctl);
			break;
		}

		if(++spotchk > 50){
			spotchk = 0;
			contentinit();
//			urlinit();
//			statsinit();
		}
	}
}

static void
doservice(char *ndir)
{
	HSPriv *hp;
	HConnect *c;
	NetConnInfo *nci;
	int ok;
	char *uri, *p;

	nci = getnetconninfo(ndir, -1);
	c = mkconnect();
	hp = mkhspriv();
	hp->remotesys = nci->rsys;
	hp->remoteserv = nci->rserv;
	myip = nci->lsys;
	port=nci->lserv; // this is real port
	c->private = hp;

	hinit(&c->hin, 0, Hread);
	hinit(&c->hout, 1, Hwrite);

	/*
	 * original comment for timeout is here:
	 * 	serve requests until a magic request.
	 * 	later requests have to come quickly.
	 * 	only works for http/1.1 or later.
	 * I don't know the meaning. the code was:
	 *	for(timeout = 15*60*1000; ; timeout = 15*1000){
	 * I replaced the code by simple one.
	 * -Kenar-
	 */

	/*	It's safe to prevent redirection loop here	*/
	for(;;){
		/* waiting requests from the clients */
		if(hparsereq(c,waittimeout *1000) <= 0) // wait timeout (in unit of milisec)
			exits(nil);
		if(ntossed > 3){ // attack or redirection loop.
			/*	3 may be sufficient for allowed toss count	*/
			syslog(0, HTTPLOG, "warning: redirection loop ?");
			hfail(c, HInternal);
			exits(nil);
		} 
		/*
		*	hparsereq() get basic parameters in the request:
		*	c->req.meth		# method "GET", "POST" ...
		*	c->req.vermaj
		*	c->req.vermin
		*	c->req.urihost		# nil, the value will be assigned by hparseheaders()
		*	c->req.uri			# uripath: path part in req. URI
		*	c->req.options		# ;params?query
		*	c->req.params		# params	got from req. URI
		*	c->req.query		# query	got from req. URI
		*
		*	scheme://urihost:port/uripath;params?query
		*	we should use real port no.
		*
		*	and also we have:
		*	c->req.user	nil
		*	c->req.path	nil
		*	these value will be assigned next.
		*	Note that there is no scheme info in the request.
		*	-Kenar-
		*/
		c->req.user = nil;
		uri = c->req.uri;
		c->req.path = uri;
		if(uri[0] == '/' && uri[1] == '~'){
			p = strchr(uri+2,'/');
			if(p)
				*p = 0;
			c->req.user = hstrdup(c, uri+2);
			if(p)
				*p = '/';
			c->req.path = p;
		}
		/*
		*	if uripath is "/~alice/foo", then
		*	c->req.user	"alice"
		*	c->req.path	"/foo"
		*	else if uripath is "/foo", then
		*	c->req.user	nil
		*	c-req.path	"/foo"
		*/


		/*	cgih must be initialized in every request */
		if(cgih)
			db_free(cgih);
		cgih = nil;
		if(errh)
			db_free(errh);
		errh = nil;

		ok = doreq(c);
		//logit(c,"DEBUG:doservice:doreq(): closeit=%d ok=%d",c->head.closeit,ok);
		if(c->head.closeit || ok < 0)
			exits(nil);
		hreqcleanup(c);
	}
}


static int
doreq(HConnect *c)
{
	char *uri, *origuri, *p, *host, *user;
	char virtualhost[100];
	static char *host0 = nil;
	static char *user0 = nil;
	char *webroot = nil;
	char buf[256];
	int status;
	int n;

	uri = c->req.uri;
	/*
	 * normal case is just file transfer
	 */
	if(cont > 0){
		logit(c, "continue: %d",cont);
		if(cont > contmax){ /* may be required for safty */
			logit(c, "continuation exceed max: %d uri=%s",cont, uri);
			return hfail(c, HInternal);
		}
		cont++;
	}

	/*
	*	Kenar complains:
	*	if(hparseheaders(c, 15*60*1000) < 0)
	*	why so long time for parsing headers?
	*	I replaced by:
	*/
	if(hparseheaders(c, parsetimeout*1000) < 0)
		exits("failed");
	/*
	*	for old client c->head.host may be nil
	*	we must supply the host name from ip value
	*/
	//logit(c,"head.host=%s", c->head.host);
	if(c->head.host == nil)
		c->head.host = hstrdup(c, getdom("ip", myip));
	if(c->head.host == nil){
		logit(c, "host: nil");
		exits("failed");
	}
	//logit(c,"head.host=%s", c->head.host);
	/*
	*	now we will have:
	*	c->head.host
	*		if HTTP Host header exists, the value will be assigned from this,
	*		if not mydom is assigned.
	*		the value is already lowered.
	*	Let ar be a host name and aichi-u.ac.jp is the dom
	*	then head.host is something like:
	*	ar
	*	ar.aichi-u.ac.jp
	*	ar.aichi-u.ac.jp:80	# some browsers send HTTP Host header with port number
	*
	*	c->head.host gives mydom info for uriconnection
	*	we can establish uriconnection
	*/
	host = c->head.host;
	n = strlen(scheme) + strlen(host) + 10;
	uriconnection = halloc(c, n);
	/*	in this stage port part exist in host	*/
	snprint(uriconnection, n, "%s://%s", scheme, host);

	/*
	*	Let's remove ":port" here from urihost
	*	although this should be done in library
	*	-Kenar-
	*/
	host = hstrdup(c, host);
	p = strchr(host,':');
	if(p)
		*p = 0;

	if(!http11(c) && !c->head.persist)
		c->head.closeit = 1;
	logit(c, "%s %s%s", c->req.meth, uri, c->req.options);
	logit(c, "header: %s", c->header);
	/*
	*	we can remove padom part from host
	*	so that we can simplify management of /sys/lib/httpd.rewrite
	*	now we don't need full name for virtual host in this file.
	*/
	if(strlen(host)>strlen(padom)){
		char *p, buf[256];
		snprint(buf, sizeof buf, ".%s", padom);
		p = host + strlen(host) - strlen(buf);
		if(strcmp(p, buf) == 0)
			*p = 0;
	}
	/*	then host is "ar in ar.aichi-u.ac.jp" for example
	*	on the other hand, host "202.250.160.40" remains as it is	*/

	//logit(c, "referer:%s", c->head.referer);

	/*
	*	We should examine the connection is really continued ?
	*	How to determine the connection should enter new name space?
	*	How to enter new name space?
	*	HTTP referer defined in HTTP/1.0 and HTTP/1.1 may be helpful.
	*	referer is http://host/path if URI is from href in html document
	*	example:	http://pc/~alice/photos/
	*	According to RFC, referer can be set not to be send for privacy
	*	But referer is nil if URI is from keyboard input,
	*	therefore referer is not acceptable for this purpose.
	*	
	*	We claim to resend if (host, user) pair has changed
	*	and close the connection
	*
	*	schemes and ports concern nothing to this problem.
	*	because different servers care for them
	*
	*	-Kenar-
	*/


	/*
	*	Problem with favicon.ico
	*
	*	Some browsers[1] always tries to get /favicon.ico
	*	in accessing directories,
	*	even if in accessing user space such as "/~alice/foo/bar/"
	*	and even if the absence of "/favicon.ico" is already have 
	*	been informed by 404 error message.
	*	This is annoying, and farther more, makes a problem in case
	*	that we should keep connection with the browser.
	*	Note that Pegasus 2.1 disconnects in constructing new name space.
	*	Pegasus 2.2 simply tries to send a file "/~alice/favicon.ico" instead of 
	*	"/favicon.ico" and keeps connection in case that Pegasus is in
	*	alice's namespace.
	*
	*	I hope the day comes such a nasty trick is wiped out.
	*
	*	[1]: Safari, Firefox for  example
	*
	*	-Kenar-
	*/
	user = c->req.user;
	//logit(c,"DEBUG: doreq: user=%s user0=%s uri=%s",user,user0,c->req.uri);
	if(user0 && user== nil){
		/* 	the Pegasus is alreay in user0 space, but the request is
		*	to the host owner's space */
		int m;
		char *p;
		m = strlen(c->req.uri) + 10;
		p = halloc(c,m);
		snprint(p,m,"/doc/~%s",c->req.uri);
		/*	if req.uri is "/favicon.ico", then p is
		*	"/doc/~/favicon.ico"		*/
		if(executable(p)){
			/* don't send executable file of real host */
			hfail(c, HNotFound);
			return 0;
		}
		sendit(c,p, HCNo);
		c->head.closeit = 0;
		return 0; //continue
	}

	if(cont){
		if(xstrcmp(host, host0) != 0 || xstrcmp(user, user0) != 0){
			/*	claim to resend and close the connection */
			c->head.closeit = 1;
			hmoved(c, c->req.uri, c->req.options);
			logit(c,"resend to:  %s%s", uriconnection, c->req.ouri);
			return -1; //exit
		}
	}
	else{
		host0 = estrdup(host);
		if(user) user0 = estrdup(user);
		/*	don't use hstrdup because memory is initialized
		*	in continued connection		*/
	}

	if((strcmp(scheme,"https") == 0 && strcmp(port,httpsport) == 0)
	|| (strcmp(scheme,"http") == 0 && strcmp(port,httpport) == 0))
		snprint(virtualhost, sizeof virtualhost, "%s://%s/", scheme, host);
	else
		snprint(virtualhost, sizeof virtualhost, "%s://%s:%s/", scheme, host, port);
	origuri = c->req.uri;
	uri = nil;
	if(user == nil)
		uri = redirect(c, rewrite, virtualhost, &webroot);
	#ifdef NOTE
	*	we have three cases of uri
	*	1. 	scheme://host/foo
	*	2. 	/foo
	*	3. 	nil
	#endif
	if(uri){
		/*	Incoming request is limited by /sys/lib/httpd.rewrite to our host	*/
		/*
		if(uri[0] == '*'){
			webroot = hstrdup(c, uri + 1);
			binding = 1;
		}
		*/
		if(webroot)
			binding = 1;
	}
	else{
		/*
		*	we can reject some faked accesses here;
		*	if uri is nil and host is not myname, we can reject.
		*	example:
		*	ar	# sysname
		*	plan9	# virtual host of ar. To be rejected.
		*	aichi-u.ac.jp	# padom
		*	now ar.aichi-u.ac.jp and plan9.aichi-u.ac.jp are reduced to
		*	ar and plan9 respectively.
		*	On the other hand
		*	www.faked.com
		*	is kept untached.
		*/
		if(strcmp(host, myname) != 0 && strcmp(host, myip) != 0){
			if(user)
				logit(c, "denied: %s/~%s", host, user);
			else
				logit(c, "denied: %s", host);
			return hfail(c, HNotFound);
		}
		/*	redirection is examined only for real host	*/
		uri = redirect(c, rewrite, origuri, &webroot);
		//logit(c,"uri=%s origuri=%s webroot=%s", uri, origuri, webroot);
		if(uri && webroot){
			/*	we just accept the incoming request	*/
			c->req.path=uri;
			if(c->req.user)
				binding = 2;
			else
				binding = 1;
		}

	}

	mnt = 0;
	if(webroot){
		#ifdef NOTE
			This case is '*' redirect in /sys/lib/httpd.rewrite such as:
			http://pegasus  */usr/someone/www
			/~alice	*/usr/alice/www
			/car		*/usr/carol/www
			/		*/usr/bob/www
		#endif
		if(!cont){
			status = bind(webroot, webmnt , MREPL);
			if(status < 0){
				logit(c,"error: bind %s %s # %r", webroot,webmnt);
				return hfail(c, HNotFound);
			}
			mnt = 1;
		}
	}
	else if(uri != nil){ /* Let the client know the new address */
		char *o;
		o = c->req.options;
		logit(c,"redirect1 to:  %s%s ntossed=%d", uri, o, ntossed);
		ntossed++;
		return hmoved(c, uri, o);
	}


	/*
	* 	Here we support such as http://ar/~alice/abc
	*	where ar must be real host and need noweb == 0.
	*	Then we bind /usr/alice/web to /usr/web/mnt
	*	and change request document to /abc
	*	This redirection is automatically done without rewrite-file
	*
	*	-Kenar-
	*/
	uri = c->req.uri;
	if(noweb == 0 && c->req.user && binding != 1){
		/*	then uri[0]=='/' && uri[1]=='~'	*/
		char *t;
		/* 	if uri is "/~alice" for example,
			let the client resend the correct uri "/~alice/"
			-Kenar-	*/
		t = strchr(uri+2, '/');
		if(!t){
			char *o;
			o = c->req.options;
			snprint(buf, sizeof buf, "%s/", uri);
			logit(c, "redirect2 to: %s%s", buf, o);
			return hmoved(c, buf, o);
		}

		if(!cont && !mnt){
			/*
			*	Pegasus 2.2 support global space such as
			*	bind $hostroot/doc/~ $web/doc/~
			*	in rc notation
			*	where
			*	hostroot=/usr/bob/www
			*	web=/usr/alice/web
			*
			*	-Kenar-
			*/

			char *user;
			char *path,*hpath;
			int n;

			n = strlen(hostroot) + 10;
			hpath = halloc(c, n);
			snprint(hpath, n, "%s/doc/~", hostroot);

			user = c->req.user;
			n = strlen("/usr//web") + strlen(user) + 10;
			path = halloc(c, n);
			snprint(path, n, "/usr/%s/web/doc/~", user);
			bind(hpath, path, MREPL); // error OK


			/*	bind $web $webmnt	*/
			snprint(path, n, "/usr/%s/web", user);
			status = bind(path, webmnt, MREPL);
			if(status == -1){
				logit(c, "error: bind %s %s # %r", path, webmnt);
				return hfail(c, HInternal);
			}
			mnt = 1;
		}

		/* 	then if c->req.uri is "/~alice/abc", the path is "/abc" 
			for the request http://ar/~alice/abc
			We need not be wary of "/~alice/../",
			it reduces to "/" as it should be.
			-Kenar-	*/
	}

	/* namespace configuration; hide everything out of webbase */
	if(!cont){
		if(addns(runas, namespace) < 0){
			logit(c, "error: addns %s # %r", namespace);
			sysfatal("can't build httpd namespace");
		}
		if(mnt){
			char nsbuf[32];
			/* 	strlen("namespace_65536") is 15	*/
			if(nsconfig(c, webbase) < 0){
				logit(c, "error: nsconfig %s # %r", webbase);
				return -1;
			}
			snprint(nsbuf, sizeof nsbuf, "/etc/namespace_%s", port);
			//logit(c,"etc/namespace: %s", nsbuf);
			if(mntok)
				addns(runas, nsbuf);
			else
				chgns(nsbuf);
		}
		cont = 1;
	}
	chdir("/");

	/*	Namespace configuration is completed.
	*	Now we are in the configured namespace.	*/

	return send(c);
}

/*	# return values
*	0,+1: OK
*	-1: NG
*	the return value should be 0 or -1
*	inconsitently used OK value. original confusion is in sentdit()
*	-Kenar-	*/
static int
send(HConnect *c)
{
	Dir *dir;
	Handle *h;
	char *p, *syspath;
	char *uri, *newuri;
	char *path;
	int n, ok;
	ulong dirmode;
	char buf[256];
	int status;
	HSPriv *hp = c->private;

	/*	Namespace configuration is completed.
	*	Now we are in the configured namespace.	*/

	if(strcmp(c->req.meth, "GET") != 0
		&& strcmp(c->req.meth, "HEAD") != 0
		&& strcmp(c->req.meth,"POST") != 0){
		logit(c, "# unallowed: %s HEAD=%s", c->req.meth, c->header);
		return hunallowed(c, "GET, HEAD, POST");
	}
	/*	NOTE
	*	PUT and DELETE method are supported by
	*		Mozilla/4.5
	*	but not supported by
	*		IE/6.0 nor Mozilla/6.2
	*	so it may be worthless to support these method
	*	-Kenar-	*/

	/*	DEBUG: monitor header */
	if(c->head.expectother || c->head.expectcont)
		logit(c, "HEAD=%s", c->header);


	/* 	RFC2068

	8.2 Message Transmission Requirements

	- An HTTP/1.1 (or later) server that receives a request from a
	   HTTP/1.0 (or earlier) client MUST NOT transmit the 100 (continue)
	   response; it SHOULD either wait for the request to be completed
	   normally (thus avoiding an interrupted request) or close the
	   connection prematurely.
	-   Upon receiving a method subject to these requirements from an
	   HTTP/1.1 (or later) client, an HTTP/1.1 (or later) server MUST either
	   respond with 100 (Continue) status and continue to read from the
	   input stream, or respond with an error status. If it responds with an
	   error status, it MAY close the transport (TCP) connection or it MAY
	   continue to read and discard the rest of the request. It MUST NOT
	   perform the requested method if it returns an error status.

	10.1.1 100 Continue

	 The client may continue with its request. This interim response is
	 used to inform the client that the initial part of the request has
	 been received and has not yet been rejected by the server. The client
	 SHOULD continue by sending the remainder of the request or, if the
	 request has already been completed, ignore this response. The server
	 MUST send a final response after the request has been completed.

	*/
	if(c->head.expectcont && http11(c))
		hokmsg(c, "100 Continue");

	if(c->head.expectother)
		return hfail(c, HExpectFail);

	path = c->req.path;
	uri = c->req.uri;

	newuri = redirect(c, "/etc/rewrite", path, nil);
	if(newuri){
		char *o;
		o = c->req.options;
		logit(c, "moved to: %s%s", newuri, o);
		ntossed++;
		return hmoved(c, newuri, o);
	}

	/*
	 * check for directory/file mismatch with trailing /,
	 * and send any redirections.
	 */
	n = strlen("/doc") + strlen(path) + strlen("/index.html") + 10;
	/* make a room to insert index.html */
	syspath = halloc(c, n);
	strcpy(syspath, "/doc");
	strcat(syspath,path);

	/*
	*	we examine if it is a directory
	*	note that the path need not exist
	*	it may be handled by handler
	*/
	if((dir = dirstat(syspath)) != nil){
		dirmode = dir->mode;
		free(dir);
		if(dirmode & DMDIR){
			/*
			*	if path does not end with '/'
			*	then let the client resent the uri with "/"
			*	Don't replace by the code such as the following
			*	because HTTP/1.0 client will not work
			*	They hate redirection
			*	-Kenar-
			*
			* 	if(path[strlen(path)-1] == '/')
			*		snprint(buf, sizeof buf, "%sindex.html", uri);
			* 	else
			*		snprint(buf, sizeof buf, "%s/index.html", uri);
			* 	return hmoved(c, buf);
			*
			*	Problem with redirection:
			*	Some browsers[1] disconnects in redirection
			*	even if the redirection is
			*	/foo => /foo/
			*	and even if "Connection: Keep-Alive" is sent to the brower
			*	Therefore be carefull in writing HTML code.
			*
			*	[1]: Safari for example
			*	Firefox does not have this problem.
			*/
			if(uri[strlen(uri)-1] != '/'){
				int n;
				char *b;
				n = strlen(uri) + 5;
				b = malloc(n);
				snprint(b, n, "%s/", uri);
				logit(c, "redirect3 to: %s", b);
				c->head.closeit = 0;
				c->head.host = nil;
				status = hmoved(c, b, c->req.options);
				if(status)
					logit(c, "#ERROR: redirect3: hmoved");
				/* 	internal error if status is -1
				*	then the connection should be freed
				*	-Kenar-	*/
				free(b);
				return status;
			}
			/*
			*	here the form of uri is "path/"
			*	that is, path ends with '/'
			*	-Kenar-
			*/
			strcat(syspath, "index.html");
		}else if((p = strchr(syspath, '\0')) > syspath && p[-1] == '/'){
			/*
			*	This condition means:
			*	path is a file and ends with '/' 
			*	for example,
			*		http://ar.aichi-u.ac.jp/index.html/
			*	Then let the client to resend
			*		http://ar.aichi-u.ac.jp/index.html
			*	-Kenar-
			*/
			*strrchr(uri, '/') = '\0';
			logit(c, "%s hmoved", uri);
			return hmoved(c, uri, c->req.options);
		}
	}


	//	--------------------------
	//	req.uri and req.path are redefined
	//	--------------------------

	/*	Let our sendit() know the contents type to assign full pathname
	*	to req.uri
	*	I dislike to replace req.uri, however sendit() uses file suffix in
	*	req.uri to determine file type.
	*	-Kenar-	*/
	path = syspath + 4; // 4 is strlen("/doc"), now `path' is path to a file
	c->req.path = path; // file path from document root

	/*
	*	Now, variables
	*		path syspath
	*	are path to file
	*/

	/*
	*	access control must be the last
	*	because it is delicate to handle last slash
	*	so full path is required to evaluate
	*	-Kenar-	*/

	if(ipcheck(c, "/etc/allow", hp->remotesys, path) < 0){
		logit(c, "ipcheck: unallowed %s %s", hp->remotesys, path);
		return hfail(c, HForbidden, path);
	}

	ok = authcheck(c, "/etc/passwd", path);
	if(ok <= 0){
		logit(c, "authcheck: unallowed %s", path);
		/* the return value should be ok because "need auth" message
		   has been sent to the client in authcheck() */
		return ok;
	}
	if(authenticated){ /* reset timer */
		waittimeout = waittimeout2;
	}

	/*	Now every redirection is finished and we have ready.
	*	requested path is already converted to absolute path.
	*	values to be used are:
	*	path: path to the file to send. where path is in httpd name space
	*	fd: open file descriptor of path
	*
	* 	CGI is supported using exechandler.
	*	The handler configuration file is /etc/handler in service space.
	*	Handler will be obtained from requested path.
	*	Handler must be exec permitted.
	*	-Kenar-
	*/

	/* we disallow files starting with `.' */
	p = strrchr(syspath,'/');
	if(p[1] == '.'){
		logit(c, "%s forbidden", uri);
		return hfail(c, HForbidden, uri);
	}

	ntossed = 0;

	/*	call exechandler	*/
	h = handler(c, "/etc/handler", path);
	if(h){
		char **argv;
		int n;
		argv = h->argv;
		for(n = 0; argv[n]; n++)
			if(strcmp(argv[n],"$target") == 0)
				argv[n] = syspath;
		if(*argv[0] != '/'){
			logit(c, "/etc/handler: %s must be absolute path", argv[0]);
			return hfail(c, HInternal);
		}
		logit(c,"%s", argv[0]);
		if(executable(argv[0])){
			if((strcmp(c->req.meth,"POST") == 0) && (acceptpost(c, posttimeout) < 0))
				return hfail(c, HReqSize);

			/* chdir to the target dir, taget is syspath in httpd space*/
			p = strrchr(syspath,'/');
			*p = 0;
			if(chdir(syspath) < 0)
				chdir("/doc");
			*p = '/';
			scriptname = argv[0];
			pathinfo = nil;
			if(syspath != scriptname)
					pathinfo = syspath;
			setcgienv(c);
			ramfs(c);
			status = execute(c,argv, h->hctl);
			unmount(nil, "/tmp");
			return status;
		}
	}

	/*	We might be prompted to prohibit to send the content
	*	of executable file, and write following codes.
	*	if(executable(syspath)){
	*		hfail(c, HForbidden);
	*		logit(c,"%s is executable, not sent", syspath);
	*		return 0; // continue
	*	}
	*	But then, Pegasus regects sending some files in netlib.
	*	Therefore I disabled codes above.
	*	-Kenar-	*/
	status = sendit(c, syspath, HCNo);
	return status; // +1: success;  -1: fail
}

/*------- Kenar's addition follows --------*/

/*	Ref: /sys/src/cmd/ndb/query.c 	*/
static	char*
sysdom(char *name)
{
	Ndb *db;
	Ndbs s;
	Ndbtuple *t;
	char buf[Ndbvlen];
	db = ndbopen("/lib/ndb/local");
	if(db == 0){
		fprint(2, "no db files\n");
		exits("no db");
	}
	t = ndbgetval(db, &s, "sys", name, "dom", buf);
	ndbclose(db);
	if(t){
		ndbfree(t);
		return estrdup(buf);
	}
	return nil;
}




/*
*	get parent domain of the server	*/
static	char *
pardom(void)
{
	char *sn, *sd, *pd;
	sn = sysname();
	sd = sysdom(sn);
	if(sd == nil)
		return nil;
	if(strlen(sd) < strlen(sn)) // something wrong
		return nil;
	pd = sd + strlen(sn) + 1; // parent dom
	return pd;
}


static void
setcgienv(HConnect *c)
{
	HSPriv *hp = c->private;
	char buf[256];
	char *cputype, *name;
	int n, ch;
	char *p;

	cputype = getenv("cputype");

	rfork(RFCENVG);

	/*	Putting header info into environment
	*	must be the first. (for security)
	*	-Kenar-	*/
	ch = *(c->hstop);
	*(c->hstop) = 0;
	setenv("HTTP_HEADER", (char*)c->header);
	hparse2env(c, (char*)c->header);
	*(c->hstop) = ch;

	setenv("DOCUMENT_ROOT", "/doc"); // CGI/1.1
	setenv("GATEWAY_INTERFACE", "CGI/1.1"); // CGI/1.1
/*	snprint(buf, sizeof buf, "%s!%s", hp->remotesys, hp->remoteserv);
	setenv("REMOTE_ADDR", buf);
*/
	setenv("PATH_INFO", pathinfo);
	p = nil;
	if(pathinfo){
		n = strlen(pathinfo) + 10;
		p = halloc(c,n);
		snprint(p, n, "/doc/%s", pathinfo);
	}
	setenv("PATH_TRANSLATED", p);
	setenv("REMOTE_ADDR", hp->remotesys); // CGI/1.1
	setenv("REMOTE_HOST", hp->remotesys); // CGI/1.1
	setenv("REQUEST_PATH", c->req.path);
	setenv("SCRIPT_NAME", scriptname); // CGI/1.1

	setenv("REQUEST_METHOD",c->req.meth); // CGI/1.1
	
	snprint(buf, sizeof buf, "%d.%d", c->req.vermaj,c->req.vermin);
	setenv("REQUEST_VERSION", buf);
	setenv("REQUEST_URI",c->req.uri);
	setenv("AUTH_TYPE", c->head.authtype); // CGI/1.1
	setenv("REMOTE_USER", c->head.authuser); // CGI/1.1
	setenv("AUTH_PASS", c->head.authpass);
	snprint(buf, sizeof buf, "%uld", c->head.contlen);
	setenv("CONTENT_LENGTH", buf); // CGI/1.1
	setenv("CONTENT_TYPE", c->head.conttype); // CGI/1.1
	setenv("REQUEST_USER", c->req.user);
	setenv("SERVER_DOMAIN", mydom);
//	setenv("SERVER_NAME", sysname());
	strncpy(buf,c->head.host,sizeof(buf));
	p = strchr(buf,':');
	if(p)
		*p = 0;
	setenv("SERVER_NAME", buf); // CGI/1.1

	setenv("SERVER_PORT", port); // CGI/1.1
	setenv("SERVER_PROTOCOL",hversion); // CGI/1.1
	setenv("SERVER_SOFTWARE", serversoft); // CGI/1.1
	setenv("QUERY_STRING", c->req.query); // CGI/1.1

	qs2env(c, c->req.query); // Pegasus Special
	setenv("target", c->req.path);
	name = strrchr(c->req.path, '/');
	setenv("name", name+1);
	setenv("query", c->req.query);
	snprint(buf, sizeof buf, "%D", time(nil));
	setenv("DATE", buf);
	setenv("home", "/doc");
	setenv("cputype", cputype);
	setenv("objtype", cputype);
}

/*
*	configure service namespace
*/
static int
nsconfig(HConnect *c, char *webbase)
{
	char *wmnt, *wdoc, *wetc, *wbin, *wmdoc, *wmetc, *wmbinrc, *wmbinobj;
	wmnt = hcat2(c, webbase,"/mnt");
	wdoc = hcat2(c, webbase, "/doc");
	wetc = hcat2(c, webbase, "/etc");
	wbin = hcat2(c, webbase,"/bin");
	wmdoc = hcat2(c, webbase, "/mnt/doc");
	wmetc = hcat2(c, webbase, "/mnt/etc");
	wmbinrc = hcat2(c, webbase, "/mnt/bin/rc");
	wmbinobj = hcat3(c, webbase, "/mnt/bin/",getenv("objtype"));
	/*	 the two unmounts are for the guard when next two binds fail.
	*	for example, /usr/web/mnt/etc is absent	*/
	unmount(nil, wdoc);
	unmount(nil, wetc);
	if(bind(wmdoc, wdoc, MREPL) < 0)
		return -1;
	bind(wmetc, wetc, MAFTER);
	bind(wmbinrc, wbin, MAFTER);
	bind(wmbinobj, wbin, MAFTER);
	unmount(nil, wmnt);
	bind(webbase,"/", MREPL);
	return 0;
}


static int
acceptpost(HConnect *c, int timeout)
{
	int fd[2];
	char *hb;
	Hio *hin;
	ChunkBuf *cb, *cb0, *cb1;
	int m,n;

	alarm(timeout*1000);

	/*	DEBUG for receiving post data over 32KB
	*	Ref: echo.c	
	print("%s 200 OK\r\n", hversion);
	print("Server: %s\r\n", serversoft);
	print("Date: %D\r\n", time(nil));
	print("Content-Type: text/plain\r\n");
	print("Connection: close\r\n");
	print("\r\n");
	hin = hbodypush(&c->hin, c->head.contlen, c->head.transenc);
	if(hin == nil){
		logit(c, "error: nil hbodypush");
		hfail(c, HInternal);
		exits("failed");
	}
	for(; hb = hreadbuf(hin, hin->pos); hin->pos = hin->stop)
		write(1, hb, hbuflen(hin, hb));
	exits(nil);
	*/

	if(c->head.contlen > maxpost){
		logit(c, "error: post data too large: %d", c->head.contlen);
		hfail(c, HReqSize, maxpost);
		exits("failed");
	}
	/*	head.contlen can be faiked. so real protection is needed	*/
	hin = hbodypush(&c->hin, c->head.contlen, c->head.transenc);
	if(hin == nil){
		logit(c, "error: nil hbodypush");
		hfail(c, HInternal);
		exits("failed");
	}
	m = 0;
	cb0 = cb = nil;
	for(; hb = hreadbuf(hin, hin->pos); hin->pos = hin->stop){
		cb1 = malloc(sizeof *cb1);
		if(cb1 == nil){
			logit(c, "error: nil malloc ChunkBuf");
			hfail(c, HInternal);
			exits("failed");
		}
		n = hbuflen(hin, hb);
		memcpy(cb1->buf, hb, n);
		cb1->size = n;
		m += n;
		if(m > maxpost){
			logit(c, "error: post data too large: %d", c->head.contlen);
			hfail(c, HReqSize, maxpost);
			exits("failed");
		}
		if(cb0 == nil)
			cb0 = cb1;
		if(cb != nil)
			cb->next = cb1;
		cb = cb1;
	}
	if(cb == nil){
		hfail(c, HBadReq);
		exits("failed");
	}
	cb->next = nil;
	alarm(0);

	/* pass the hb to exec() */
	pipe(fd);
	switch(fork()){
	case -1:
		logit(c, "error: fork # %r");
		hfail(c, HInternal);
		return -1;
	case 0:
		/*	Let the child exit, because I would like to let CGI
		*	succeed to this pid
		*	-Kenar-	*/
		close(0);
		close(fd[0]);
		dup(fd[1], 1);
		close(fd[1]);
		/*	POST data have been received in ChunkBuf cb0,
		*	so we pass the data to the child process */

		cb = cb0;
		while(cb){
			m=write(1, cb->buf, cb->size);
			if(m != cb->size){
				logit(c, "error: write # %r");
				hfail(c, HInternal);
			}
			cb = cb->next;
		}
		exits(nil);
	default:
		close(fd[1]);
		dup(fd[0], 0);
		close(fd[0]);
		/* remove ChunkBuf pointed by cb0 */
		cb = cb0;
		while(cb){
			cb1 = cb->next;
			free(cb);
			cb = cb1;
		}
		break;
	}
	return 0;
}
			
static int
execute(HConnect *c, char *hargv[], int hctl)
{
	char *p, *q;
	char *params;
	int k, n, hargc;
	char **argv;
	char *path;

	params = c->req.params;
	/*	params points "alice;bob"
	*	In creating argv for exec call,
	*	we should allocate memory for argv.
	*	argv[0] must be reserved for basename of exec file
	*	argv[n-1] must be nil where n is num of elements
	*	so, I use ";alice;bob;" to apply getfield.
	*	-Kenar-		*/

	//logit(c, "DEBUG: params %s",params);

	if(params == nil)
		argv = hargv;
	else{
		for(n = 0; hargv[n]; n++);
		hargc = n;
			
		n = strlen(";") + strlen(params) + 1 + UTFmax;
		q = halloc(c, n);
		snprint(q, n, "%s;", params);
		params = q;
		n = 0;
		for(p = params; *p; p++)
			if(*p == ';') n++;
		argv = halloc(c, (hargc + n+1) * sizeof(char*));
		n = getfields(params, &argv[hargc], n+1, 0, ";");
		for(k = 0; k < hargc; k++)
			argv[k] = hargv[k];
		/* don't forget to hurlunesc */
		for(k = hargc; k < hargc + n -1; k++)
			argv[k] = hurlunesc(c, argv[k]);
		/* it is not good idea to trim extra spaces */
		argv[hargc + n-1] = nil;
	}
	path = argv[0];
	p = strrchr(argv[0],'/');
	argv[0] = p + 1;

	/*	if it takes so much time to complete, kill the process
	*	-Kenar-	*/
	return execsafely(c, path, argv, hctl);
}


/*	This code intends to get exit status from the child	*/
static int
execsafely(HConnect *c, char *path, char **argv, int hctl)
{
	int fd[2], fd0, n;
	Biobuf ib;
	int status;
	char msg[ERRMAX];
	char *exitmsg;
	char *resheader=nil;
	Hio *hout;
	char *p;

	/* 	wait child to finish.
	*	If it takes too much time, kill the child and get message  */
	if(pipe(fd)<0)
		hfail(c, HInternal);

	switch(cpid = rfork(RFFDG|RFPROC|RFNOTEG|RFREND|RFNAMEG)){ /* assign = */
	/*	RFNAMEG: the new process inherits a copy of the parent's name space.
	*	RFNAMEG is required so that nomnt() does not make effect parent namespace	*/
	case -1:
		logit(c, "error: fork # %r");
		hfail(c, HInternal);
		return -1;
	case 0: // child: read stdin, and write to stdout
		dup(fd[1], 1); // the stdout is redirected to fd
		dup(fd[1], 2); // the stderr is also redirected to fd
		close(fd[1]);
		if(!mntok)
			nomount(c);
		exec(path, argv);
		/* not reached */
		logit(c,"error: exec %s # %r", path);
		exits("500 Internal Server Error");// don't return
		break;
	default:// parent: read data from the child
		close(fd[1]);
		break;
	}
	atnotify(notefun,1);
	alarm(cgitimeout*1000);
	notestatus = 1;

	Binit(&ib, fd[0], OREAD);
	fd0 = fd[0];
	hout = &c->hout;

	status = 0;
	if(hctl == HCFull){
		char *p;
		int n;

		Binit(&ib, fd0, OREAD);
		while((p = Brdstr(&ib,'\n',1)) && *p){
			n = strlen(p); // then n > 0 because *p != 0
			if(p[n - 1] == '\r'){
				p[n - 1] == 0;
				if(n == 1)
					break;
			}
			hprint(hout, "%s\r\n", p);
		}
		hprint(hout,"\r\n");
		while((p = Brdstr(&ib,'\n',1)))
			hprint(hout,"%s\n", p);
		logit(c, "done %s", path);
		return hflush(hout);
	}
	if(hctl == HCPlus){
		Biobuf out;
		char *av[2];
		int n;
		int fd1;

		while((p = Brdstr(&ib,'\n',1)) && *p){
			n = strlen(p);
			if(n > 0 && p[n-1] == '\r'){
				n--;
				p[n] = 0;
			}
			if(n == 0)
				break;
			n = getfields(p,av,2,1," ");
			if(n != 2){
				logit(c, "# error: CGI format");
				return hfail(c, HInternal);
			}
			av[1] = trim(av[1]);
			if(strcmp("Content-Length:",av[0]) != 0)
				db_put(&cgih,av[0],av[1]);
		}

		fd1 = create("/tmp/...", OWRITE, 0600);
		if(fd1 < 0)
			return hfail(c, HInternal);
		n = 0;
		while((p = Brdstr(&ib,'\n',1))){
			fprint(fd1,"%s\n", p);
			n += strlen(p);
		}
		close(fd1);
		if(n == 0){
			p = db_get(cgih,"Status:");
			if(p && strncmp(p,"200 ", 4) != 0 && strncmp(p,"301 ", 4) != 0)
				return hfail0(c, p, nil, nil);
			p = db_get(cgih,"Location:");
			if(p)
				return hmoved(c, p, nil);
		}
		p = db_get(cgih,"Location:");
		if(p)
			db_put(&cgih,"Status:", "301 Moved Permanently");
	}
	else if(hctl == HCNo){
		int fd1;

		fd1 = create("/tmp/...", OWRITE, 0600);
		if(fd1 < 0)
			return hfail(c, HInternal);
		while((p = Brdstr(&ib,'\n',1)))
			fprint(fd1,"%s\n", p);
		close(fd1);
		status = hflush(hout);
	}
	else{
		logit(c,"# invalid hctl");
		return hfail(c, HInternal);
	}

	status = waitfor(cpid, msg);
	alarm(0);
	atnotify(notefun,0);
	close(fd0);
	if(notestatus != 1){
		logit(c,"alarm done to pid=%d", cpid);
		return hfail(c, HCgiTimeout);
	}

	/*	example of msg:
	*	httpd 1653: 500 Internal Server Error
	*	therefore we must extract what we want	*/
	exitmsg = nil;
	if(msg[0] != 0){
		exitmsg = strchr(msg, ':');
		if(exitmsg && exitmsg[1] == ' ')
			exitmsg += 2;
		else
			exitmsg = msg;
		logit(c,"child %d exited with msg [%s]", cpid, exitmsg);
		if(!httpstatus(exitmsg))
			return hfail(c,HCGIErr,exitmsg);
		c->head.closeit = getresheader(c, exitmsg, &resheader);
		if(strcmp(resheader,"200 OK") != 0)
			return hfail0(c,exitmsg,nil,nil);
	}

	p = db_get(cgih, "Connection:");
	if(p){
		c->head.closeit = 0;
		if(cistrcmp(p,"close") == 0)
			c->head.closeit = 1;
		db_del(&cgih, "Connection:");
	}

	status = sendit(c, "/tmp/...", hctl);

	if(exitmsg)
		logit(c, "done %s # [%s]",path, exitmsg);
	else
		logit(c, "done %s", path);
	return status;
}

static int
ramfs(HConnect *c)
{
	int pid, status;
	char msg[ERRMAX];
	/*	ramfs service to the child	*/
	switch(pid = rfork(RFFDG|RFPROC)){ /* assign = */
	case -1:
		logit(c, "error: fork # %r");
		hfail(c, HInternal);
		return -1;
	case 0:
		close(0);
		close(1);
		close(2);// prevent "ramfs: unmount read" message
		execl("/bin/ramfs","ramfs","-p",0);
		/* not reached */
		logit(c, "error: unable to execl /bin/ramfs %r");
		exits(nil);
	default:
		break;
	}
	
	/* ensure "/bin/ramfs" finished */
	status = waitfor(pid, msg);
	if(status < 0){
		logit(c,"error: /bin/ramfs: no children to wait %d", pid);
		return -1;
	}
	return 0;
}

static int
nomount(HConnect *c)
{
	int status;
	status = rfork(RFNOMNT);
	if(status < 0){
		logit(c,"error: rfork nomnt # %r");
		return hfail(c, HInternal);
	}
	return 0;
}

/*	mime parse is not easy.
*	and probably meaningless to put the info int env.
*	so I chose a simple way
*	I parsed only
*		key:val
*	pairs, and
*		putenv(key, val)
*	-Kenar-	*/
static int
hparse2env(HConnect *c, char *head)
{
	char *p,*s;
	char **buf;
	char *kv[2];
	char buf1[64];
	int k,n,m;

	p = hstrdup(c, head);
	/* skip a line */
	p = strchr(p, '\n');
	if(!p)
		return -1;
	p++;
	n = 0;
	for(s = p; *s; s++)
		if(*s == '\n') n++;
	buf = halloc(c, (n + 1) * sizeof(char*));
	n = getfields(p, buf, n + 1, 0, "\n");
	
	/* strip the trailing CR */
	for(k = 0; k < n; k++){
		p = strchr(buf[k], '\r');
		if(p)
			*p = 0;
	}

	/*	get
	*		key: val
	*	and trim, then putenv	*/
	for(k = 0; k < n; k++){
		m = getfields(buf[k], kv, 2, 0, ":");
		if(m != 2)
			continue;
		for(p = kv[0]; *p; p++)
			*p = toupper(*p);
		for(p = kv[0]; *p; p++)
			if(*p == '-') *p = '_';
		snprint(buf1, sizeof buf1, "HTTP_%s", trim(kv[0]));
		setenv(buf1, trim(kv[1]));
	}
	return 0;
}


/*	"/proc" is required for postnote	*/

static int
notefun(void *a, char *msg)
{
	USED(a);
 	postnote(PNGROUP, cpid, msg);
	if(strcmp(msg, "alarm") == 0){
		notestatus = 0;
		return 1;
	}
	return 0;
}

/*
*	The format of "/etc/allow" is shown using example:
*	/lecture/for/our/students
*		192.168.[1-8].*
*		...
*	path
*		...
*	where ip is given by shell style pattern matching.
*
*	restrictions := restriction restriction ...
*	restriction := pathline iplines
*	pathline := path
*	iplines := ipline ipline ..
*	ipline := spaces ippattern ippattern ..
*
*	# return values
*	0: the IP is allowed
*	-1: the IP is disallowed
*/
static int
ipcheck(HConnect *c, char *file, char *ip, char *path)
{
	char *p, *line, *s;
	int ok, check;//, m;
	Biobuf *b;
	char word[64];

	//logit(c,"ipcheck: ip=%s", ip);

	b = Bopen(file, OREAD);
	if(b){
		ok = 1;
		check = 0;
		while((line = Brdline(b, '\n')) != nil){
			line[Blinelen(b)-1] = 0;
			s = strchr(line, '#');
			if(s != nil)
				*s = '\0'; 	/* chop comment */
			/* skip null line */
			if(*skip(line," \t") == 0)
				continue;

			if(strchr(" \t", line[0])){	// ip pattern of lpath
				if(check == 0) // skip the ip pattern
					continue;
				if(ok)
					continue;
				/* 	how can I get single pattern ?
				*	ip pattern has no space	OK */
				while(line = getword(line, word, sizeof word)){ // asign =
					//logit(c,"ipcheck: word=%s", word);
					if(word[0] == '~' && match(ip, word+1) == 0){
						ok = 1;
						break;
					}
					if(match(ip, word)){
						ok = 1;
						break;
					}
				}
				if(!ok) // examine more ip line
					continue;
			} else {	// path
				/* first we must check previous analysys */
				if(check && !ok)
					break;
				/* start new path line */
				p = strpbrk(line," \t");
				if(p)
					*p = 0;
				/* 	check some human error, in error put some
				*	guards and let that know. */
				if(line[0] != '/'){
					logit(c, "warning: path in /etc/allow must begin with '/': %s",line);
					path++;
				}
				if(deeper(path, line)){
					//logit(c, "deeper: %s %s",path,line);
					/* 	then we must confirm
					*	the client IP is allowed to see our data */
					check = 1;
					ok = 0;
				}
				else{
					check = 0;
					ok = 1;
				}
			}
		}
		Bterm(b);
		return  ok - 1;
	}
	return 0;
}

/*
*	uriloop(pat, rep, uri) returns:
*	 1 if rep makes loop from the pat
*	 0 if not
*
*	uri is incoming uri of common form
*		scheme://host[:port]
*	pat is one of the form:
*		scheme://host[:port][/path]
*		/path
*	rep is one of the form:
*		scheme://hostdom[:port][/path][;others]
*		/path[;others]
*	scheme and host of pat and rep is assumed to be lower.
*	port part is assumed to be eliminated if it is:
*		80 for http
*		443 for https
*/

static int
uriloop(char *pat, char *rep, char *uri)
{
	int c;
	char *p, *q;
	char *repport;
	//syslog(0,HTTPLOG,"uriloop: %s %s %s", pat, rep, uri);
	/*	comparison under common form is required	*/
	repport = nil;
	p = rep;
	if(strncmp(p, "http://", 7) == 0){
		p += 7;
		repport = httpport;
	}
	else if(strncmp(p, "https://", 8) == 0){
		p += 8;
		repport = httpsport;
	}
	if(repport == nil) // the rep is not this host, not looped
		return 0;
	/*	then p points host part	*/
	q = skipto(p, "/");
	c = *q;
	if(*q == '/')
		*q = 0;
	/*	then p is "ar.aichi-u.ac.jp" or "ar.aichi-u.ac.jp:8080" for example	*/
	if(strcmp(uri, rep) != 0){// the rep is not this host, not looped
		*q = c;
		return 0;
	}
	*q = c; // then q is path part that may be "/somepath" or ""
	if(*q) // then *q == '/'
		return deeper(q, pat);
	// then pat has not path part, we can asume 
	return (strlen(pat) == 1);
}

#ifdef NOTE
*	path redirection for users and virtual hosts
*	this redirection is used after namespace configuration
*
*	example:
*	request is http://ar/~alice/photo/index.html;args?query
*	then,
*	path is "/photo/index.html"
*	c->req.user is "alice"
*	c->head.host is "ar"
*	
*	How to design rewrite file ?
*	A line consists two fields except comment:
*		pattern	redirection
*
*	We support by this file:
*	1. local redirection
*	2. non local redirection
*	3. port redirection
*	4. scheme redirection
*	5. binding
*	We introduce special charactor '*'
*	then types of second fields are:
*		/path		# local redirection
*		*/path		# binding
*		scheme://host/path		# non local redirection
*		scheme://host:port/path	# non local redirection with port
*	- server must know current scheme and port to prevent redirection loop
*	but how to know current scheme?
*	- we have only two scheme that is accepted by httpd: http and https
*	the latter has certificate.
*	- it is enought to compare host address and port to prevent redirection loop.
*	scheme is needless.
*
*	Example redirection file that can be consulted by both http and https
*		http://vhost	*/vhostwebroot
*		/netlib	ftp://somehost			# note not only http or https
*		/oldplace	/newplace			# local redirection
*
*	Convert second field:
*		http://host	->	http://host:80
*	in checking redirection loop.
*	Pre-assign my scheme host port string such as:
*		http://ar:80
*	Then compare.
*	It is safe to remain scheme part so that we need not talk except http and https.
*
*	How should I do in case:
*		/		http://ar.aichi-u.ac.jp
*		/		http://202.250.160.40
*		/private	https://ar.aichi-u.ac.jp/private
*	These make a redirection loop if httpd in running on ar.aichi-u.ac.jp.
*	Hmm... "//*" is not good idea...
*	Second field requires full address.
*	Example redirection file that can be consulted by both http and https
*		http://vhost	*/vhostwebroot
*		/private	https://ar.aichi-u.ac.jp/private	# this is skipped by https
*		/special	http://ar.aichi-u.ac.jp:8080/special	# this is skipped by 8080 http
*		/netlib	ftp://somehost			# note not only http or https
*		/oldplace	/newplace			# local redirection
*		/		http://ar.aichi-u.ac.jp/	# this is skipped by http
*	
*
*	# return values
*	nil: not redirected
*	else: new uri
*	-Kenar-
#endif

static char*
redirect(HConnect *c, char *file, char *path, char **pwebroot)
{
	char *pat, *rep, *line, *s, *buf;
	Biobuf *b;
	char *field[2];
	int bufsize;
	char myuri[64]; // "http://ar.aichi-u.ac.jp/" ...  # scheme://host[:port]

	b = Bopen(file, OREAD);
	if(b){
		pat = nil;
		rep = nil;
		while((line = Brdline(b, '\n')) != nil){
			line[Blinelen(b)-1] = 0;
			s = strchr(line, '#');
			if(s != nil)
				*s = '\0'; 	/* chop comment */
			if(getfields(line, field, 2, 1, " \t") != 2)
				continue;
			if(deeper(path, field[0])){
				pat = field[0];
				rep = field[1];
				//logit(c, "redirect: deeper: %s %s", pat,rep);
				break;
			}
		}
		if(rep == nil){ // never happen
			Bterm(b);
			return nil;
		}
		if(strcmp(pat,"/") == 0 && strncmp(path, "/~", 2) == 0){
			hostroot = hstrdup(c,rep+1);
			Bterm(b);
			return nil;
		}
		if(rep[0] == '*'){
			rep++;
			/*	we cut off trailing '/'	*/
			if(*(rep + strlen(rep) - 1) == '/')
				*(rep + strlen(rep) - 1) = 0;
			//rep = hstrdup(c,rep);	
			if(pwebroot)
				*pwebroot = hstrdup(c,rep);
			#ifdef NOTE
			*	In rewrite file we allow
			*	/car		*/usr/carol/www
			*	/~carol	*/usr/carol/www
			*	/		*/usr/bob/www
			* 	Then path="/car/somepath" should return "/somepath"
			#endif
			/*	cut off trailing "/" for safety. this is especially true for pat="/"	*/
			if(*(pat + strlen(pat) - 1) == '/')
				*(pat + strlen(pat) - 1) = 0;
			rep = hstrdup(c, path + strlen(pat));
		}
		else {// rep is "/...." or "scheme://...." 
			if((strcmp(scheme,"http") == 0 && strcmp(port,httpport) == 0)
			|| (strcmp(scheme,"https") == 0 && strcmp(port,httpsport) == 0))
				snprint(myuri, sizeof myuri, "%s://%s", scheme, c->head.host);
			else
				snprint(myuri, sizeof myuri, "%s://%s:%s", scheme, c->head.host,  port);
			if(uriloop(pat, rep, myuri)){
				//logit(c, "redirect: uriloop: 1");
				return nil;
			}
			bufsize = strlen(rep) + strlen(path) - strlen(pat) + 1 + UTFmax;
			buf = halloc(c, bufsize);
			s = path + strlen(pat);
			if(*s == '/' || *(rep+strlen(rep) -1) == '/' || *s == 0)
				snprint(buf, bufsize, "%s%s", rep, s);
			else
				snprint(buf, bufsize, "%s/%s", rep, s);
			rep = buf;
			if(pwebroot)
				*pwebroot = nil;
		}

		if(strcmp(pat,"/") == 0)
			hostroot = rep;
		else{
			while((line = Brdline(b, '\n')) != nil){
				line[Blinelen(b)-1] = 0;
				s = strchr(line, '#');
				if(s != nil)
					*s = '\0'; 	/* chop comment */
				if(getfields(line, field, 2, 1, " \t") != 2)
					continue;
				if(strcmp(pat,"/") == 0)
					break;
			}
			hostroot = strdup(field[2] + 1); // for example /usr/bob/www
		}
		Bterm(b);
		return rep;
	}
	return nil;
}

/*	exec handler
*
*	# data format:
*	path	mimetype	hctl	path	args ...
*
*	# example:
*	.dx_html	text/html	0	/bin/dx
*
*	# return values
*	nil: need not to handle
*	else: handling command line
*	-Kenar-	*/
static Handle*
handler(HConnect *c, char *file, char *path)
{
	char *line, *s, *p;
	Biobuf *b;
	char *field[16+1], **v;
	Handle *h;
	int n, ok;
	if(!path)
		return nil;
	b = Bopen(file, OREAD);
	if(b){
		ok = 0;
		p = halloc(c, 256);
		while((line = Brdline(b, '\n')) != nil){
			line[Blinelen(b)-1] = 0;
			s = strchr(line, '#');
			if(s != nil)
				*s = '\0'; 	/* chop comment */
			strncpy(p, line, 256);
			n = tokenize(p, field, 16);
			if(n < 4)
				continue;
			if(match(path, field[0])){
				ok = 1;
				field[n] = nil;
				break;
			}
		}
		Bterm(b);
		if(!ok)
			return nil;
		h = halloc(c, sizeof(Handle));

		if(strcmp(field[1],"-") != 0)
			db_put(&cgih, "Content-Type:",field[1]);
//		h->mimetype = field[1];
		
		switch(*field[2]){
		case '0': h->hctl = HCNo;
			break;
		case '+': h->hctl = HCPlus;
			break;
		case '1': h->hctl = HCFull;
			break;
		default: h->hctl = HCNo;
		}

		/* we must have a room for additional $target at the end */
		v = halloc(c, n * sizeof(char*));
		for(n = 3; field[n]; n++)
			v[n - 3] = field[n];
		v[n -3] = nil;
		h->argv = v;
		return h;
	}
	return nil;
}

/*	I hate interface like:
*	hcat(c, arg1, arg2, nil)
*	because I forget last nil	*/
static char *
hcat2(HConnect *c, char *a1, char *a2)
{
	int m;
	char *b;
	m = strlen(a1) + strlen(a2) + 1 + UTFmax;
	b = halloc(c, m);
	strcpy(b, a1);
	strcat(b, a2);
	return b;
}

/*	I hate interface like:
*	hcat(c, arg1, arg2, nil)
*	because I forget last nil	*/
static char *
hcat3(HConnect *c, char *a1, char *a2, char *a3)
{
	int m;
	char *b;
	m = strlen(a1) + strlen(a2) + 1 + UTFmax;
	b = halloc(c, m);
	strcpy(b, a1);
	strcat(b, a2);
	strcat(b, a3);
	return b;
}

/*
*	we have two cases of uri
*	scheme://host/foo
*	/foo		# where should I go?
*
*	Note: in last case, hmoved() cares for putting scheme://host
*	but the original hmoved() can care only http.
*	must be fixed!
*	scheme should be kept in redirection.
*
*	It's better to close connection if destination is not
*	this name space.
*	uri:
*	"/~who/path"
*	"/path"
*/
char *
addoptions(HConnect *c, char *uri)
{
	char *new;
	int n;
	char *p, *u;
	char *option;
	option = c->req.options;
	if(option == nil)
		option = "";
	u = nil; // user in uri
	if(uri[0] == '/'){
		if(uri[1] == '~'){
			p = strchr(uri+2, '/');
			if(p)
				*p = 0;
			u = hstrdup(c, uri+2);
			if(p)
				*p = '/';
		}
		if(xstrcmp(c->req.user, u) != 0)
			c->head.closeit = 1;
	}
	else
		c->head.closeit = 1;
	n = strlen(uri) + strlen(option) + 1 + UTFmax;
	new = halloc(c, n);
	snprint(new, n, "%s%s", uri, option);
	return new;
}

int
httpstatus(char *status)
{
	if(strlen(status) > 3
	&& isdigit(status[0]) &&  isdigit(status[1]) &&  isdigit(status[2])
	&& status[3] == ' ')
		return 1;
	return 0;
}


/*
*	getresheaders() get HTTP headers.
*	status is exit-message of a CGI program
*	If stasus is nil or "" , it means normal exit.
*	CGI program should end with exit-message of some HTTP
*	response rule if it does not end normally.
*	However my old CGI programs are coded in loose manner;
*	they exits with groundless message.
*	Therefore getresheader should handle these exit-message
*	so that it is compatible with older version of Pegasus:
*	if status is not HTTP response format we should assign
*		status = "200 OK"
*		closeit = 1
*/
static int
getresheader(HConnect *c, char *status, char **resheader)
{
	Hio *hout = &c->hout;
	char *p;
	int f=0; // f = 1 is status is HTTP response message
	int closeit = 0;
	if(status == nil || strlen(status) < 3){
		*resheader = "200 OK";
		return 0;
	}
	if(httpstatus(status))
		f = 1;
	if(strlen(status) > 3 && !isspace(status[3]))
		f = 0;
	if(f == 0){
		status = "500 Internal Error";
		closeit = 1;
	}
	else{
		status = hstrdup(c, status);
		p = strchr(status, '#');
		if(p){
			*p++ = 0;
			p = trim(p);
			if(cistrcmp(p, "close") == 0)
				closeit = 1;
			else if(cistrcmp(p, "keep") == 0)
				closeit = 0;
			status = trim(status);
		}
		else{
			if(status[0] == '1' || status[0] == '2')
				closeit = 0;
			else
				closeit = 1;
		}
	}
	*resheader = status;
	return closeit;
}


/*	we are in emergency.
*	therefore we send a simple message to the client. sorry	*/
int
creject(int ctl, char *ndir)
{
	int data;
	int status = 0;
	char *msg ="\
HTTP/1.1 503 Service Unavailable\r\n\
Server: Plan9\r\n\
Content-Type: text/plain\r\n\
Content-Length: 5\r\n\
Connection: close\r\n\
\r\n\
Busy\n";

	switch(rfork(RFFDG|RFPROC|RFNOWAIT|RFNAMEG)){
	case -1:
		status = -1;
		break;
	case 0:
		/*
		 *  see if we know the service requested
		*
		*	What is the first argument of accept should be ?
		*	man dial(2) says, it is nctl
		*	however /sys/src/cmd/ip/httpd/httpd.c says, ctl
		 */
		data = accept(ctl, ndir);
		if(data < 0){
			syslog(0, HTTPLOG, "error: can't open %s/data: %r", ndir);
			exits(nil);
		}
		write(data, msg, strlen(msg));
		close(data);
		exits(nil);
	default:
		break;
	}
	return status;
}
